/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 *
 * 04.07.90 kk  replaced begin/end critical by dummy procedures
 * 23.07.90 kk  added unsigned to the generation of random_id and permission
 *              tables
 */

#include "sysconfig.h"
#include "np.h"


/*
 * Generate functions to be executed at the entry to a stub,
 * thereby defining a critical region.
 */

server_begin_critical()
{
}


/*
 * Generate functions to be executed at the exit from a stub,
 * thereby defining a critical region.
 */

server_end_critical()
{
}


/*
 * Make the server export binding procedures according to what the user has
 * specified in the connection statements.
 */

server_export( lib_name, module_name, protocol_suffix )
	char *lib_name;
	char *module_name;
	char *protocol_suffix;
{
	list connlist;

	/* Loop through the list of connection procedures and generate  */
	/* them.                                                        */
	for( connlist = c_server.s_connection; connlist != NIL; connlist = cdr( connlist ) ) {
		/* Generate one connection procedure stub.              */
		server_oneexport( car( connlist ), lib_name, module_name, protocol_suffix );
	}
}


/*
 * Make one server export binding procedure according to what the user has
 * specified in the connection argument list for that procedure.
 */

server_oneexport( oneconn, lib_name, module_name, protocol_suffix )
	list oneconn;
	char *lib_name;
	char *module_name;
	char *protocol_suffix;
{
	struct object *name, *type;
	struct ie_object *ie;
	list connsymtab;
	int i;

	/* Get a pointer to the name of the procedure.                  */
	name = ocar( car( oneconn ) );

	/* Get a pointer to the type tree.                              */
	type = ocdr( car( oneconn ) );

	/* Get a pointer to the symbol table.                           */
	connsymtab = cdr( oneconn );

	/* Generate the procedure header according to the specs.        */
	gen_phead( server, "int", "", "", make_prefix( name->o_name ), module_name, type );
	fprintf( server,
"\tint service_entry_id;\n\n\
\t_%s_export = -1;\n",
		module_name );

	/* Begin critical section.                                      */
	server_begin_critical();
	fprintf( server,
"\tservice_entry_id = %s(", lib_name );

	/* Loop through the import/export table to find out if any of   */
	/* the parameters are specified.                                */
	for( i = 1, ie = ie_table; ie->ie_name != 0; ++i, ++ie ) {
		name = make_symbol( ie->ie_name );
		if( lookup( connsymtab, name ) == (struct object *)0 ) {
			switch( ie->ie_type ) {
				case IE_SERVICE:
					fprintf( server, "\"%s\", ", module_name );
					break;

				case IE_HOSTNAME:
					if( c_server.s_hostname ) {
						fprintf( server, "\"%s\", ", c_server.s_hostname );
					} else {
						fprintf( server, "%s, ", "NC_E_STRING" );
					}
					break;

				case IE_PORTNUMBER:
					fprintf( server, "%d, ", c_server.s_portnumber );
					break;

				case IE_PROTECTION:
					fprintf( server, "%s, ", c_server.s_protection );
					break;

				case IE_EXECUTION:
					fprintf( server, "%s, ", c_server.s_execution );
					break;

				case IE_LOAD:
					fprintf( server, "%s, ", c_server.s_load );
					break;

				case IE_FILENAME:
					fprintf( server, "%s, ", c_server.s_execfile );
					break;

				case IE_DIRECTORY:
					fprintf( server, "%s, ", c_server.s_directory );
					break;

				case IE_DIRHOST:
					if( c_directory.d_hostname ) {
						fprintf( server, "\"%s\", ", c_directory.d_hostname );
					} else {
						fprintf( server, "%s, ", "NC_E_STRING" );
					}
					break;

				case IE_DIRPORT:
					fprintf( server, "%d, ", c_directory.d_portnumber );
					break;

				case IE_PROTOCOL:
					fprintf( server, "%s%s, ", c_transport.tp_name, protocol_suffix );
					break;

				case IE_LOADPROTOCOL:
					fprintf( server, "%s%s, ", c_transport.tp_load, protocol_suffix );
					break;

				case IE_DIRPROTOCOL:
					fprintf( server, "%s%s, ", c_transport.tp_directory, protocol_suffix );
					break;

				case IE_TOABSOLUTE:
					fprintf( server, "%d, ", c_server.s_toabsolute );
					break;

				case IE_TORELATIVE:
					fprintf( server, "%d, ", c_server.s_torelative );
					break;

				default:
					error(FATAL, "illegal parameter class in server_export");
					/* NOTREACHED */
			}
		} else {
			fprintf( server, "%s, ", name->o_name );
		}

		/* Free the allocated space.                            */
		free( name->o_name );
		free( name );

		/* Format the output in a nice way.                     */
		if( i == 5 ) {
			fprintf( server, "\n\t\t" );
			i = 0;
		}
	}

	/* Complete the procedure call.                                 */
	fprintf( server, "0%o, %s, %s, %s, %d,\n\
\t\t_%s_procedures, sizeof(_%s_procedures)/sizeof(int (*)()),\n\
\t\t_%s_permission, _%s_random, %d, &_%s_export);\n",
		c_server.s_permission, c_server.s_idffile, c_server.s_parameters,
		c_server.s_policy, c_server.s_maxconn,
		module_name, module_name, module_name,
		module_name, hash_service( module_name ), module_name );

	/* End critical section.                                        */
	server_end_critical();

	/* Terminate the procedure.                                     */
	fprintf( server,
"\treturn (service_entry_id);\n\
}\n" );
}


/*
 * Make the server free binding procedure according to what the user has
 * specified in the connection argument list.
 */

server_free( lib_name, module_name )
	char *lib_name;
	char *module_name;
{
	/* Make the procedure header.                                   */
	fprintf(server,
"\n\
int\n\
free_%s(service_entry_id)\n\
\tint service_entry_id;\n\
{\n\
\tint status;\n\n",
	module_name);

	/* Begin critical section.                                      */
	server_begin_critical();
	fprintf(server,
"\tstatus = %s(service_entry_id);\n\
\t_%s_export = -1;\n",
		lib_name, module_name );

	/* End critical section.                                        */
	server_end_critical();
	fprintf(server,
"\treturn (status);\n\
}\n");
}


/*
 * Make the server bind procedure according to what the user has
 * specified in the connection argument list.
 */

server_bindto( lib_name, module_name )
	char *lib_name;
	char *module_name;
{
	/* Make the procedure header.                                   */
	fprintf(server,
"\n\
bind_%s(service_entry_id)\n\
\tint service_entry_id;\n\
{\n\
\t_%s_export = service_entry_id;\n\
}\n",
		module_name, module_name);
}


/*
 * Generate server procedure pointer table and the corresponding table
 * to check if the interface is consistent.
 */

server_table( module_name )
	char *module_name;
{
	list p;
	struct object *t;

	/* Generate the procedure call table.                           */
	fprintf(server,
"\n\
static int (*_%s_procedures[])() = {\n",
		module_name);

	/* List all the procedures declared in the module.              */
	/* We rely on the fact that declare() builds its list in order. */
	for (p = procedures; p != NIL; p = cdr(p)) {
		fprintf(server,
"\tserver_%s",
			ocar(car(p))->o_name);
		if (cdr(p) != NIL)
			fprintf(server, ",\n");
		else
			fprintf(server, "\n");
	}
	fprintf(server, "};\n");

	/* Generate the consistency check table.                        */
	fprintf(server,
"\n\
static unsigned long _%s_random[] = {\n",
		module_name);

	/* List all random numbers assigned to procedures in the module.*/
	/* We rely on the fact that declare() builds its list in order. */
	for (p = procedures; p != NIL; p = cdr(p)) {
		t = ocdr(car(p));
		fprintf(server,
"\t%ld",
			t->t_typecheck);
		if (cdr(p) != NIL)
			fprintf(server, ",\n");
		else
			fprintf(server, "\n");
	}
	fprintf(server, "};\n");

	/* Generate the permission control table.                       */
	fprintf(server,
"\n\
static unsigned short _%s_permission[] = {\n",
		module_name);

	/* List all permissions assigned to procedures in the module.   */
	/* We rely on the fact that declare() builds its list in order. */
	for (p = procedures; p != NIL; p = cdr(p)) {
		t = ocdr(car(p));
		fprintf(server,
"\t%d",
			ocar(t->t_permission)->o_value);
		if (cdr(p) != NIL)
			fprintf(server, ",\n");
		else
			fprintf(server, "\n");
	}
	fprintf(server, "};\n");
}
