/*
 * Copyright (C) 1986 by CERN.
 *
 * 19.05.90     First written by Kris Kostro (nc version 2)
 * LP - 23.05.2000 - fixed time.h, types.h, unistd.h inclusion
 * LP - 23.08.2000 - removed fprintf in line 124 because cd->ncc_name
 *                   not initialized crashes the server when nc_getcall()
 *                   fails (bus error)
 * LP - 28.08.2000 - log date and time on nc_getcall() failure
 * CS - 07.09.2000 - cleand up some conditional compilations 
 */

#include "sysconfig.h"

#include <stdlib.h>
#ifdef OS9
  #include <time.h>
  #include <types.h>
  #include "select.h"     /* LP - new proto */
  #include "chlogs.h"     /* LP/CS-logging facility */
#else
  #include <sys/time.h>
  #include <sys/types.h>  /* LP - will include select.h too */
  #include <unistd.h>
#endif


#include "errtype.h"
#include "ncc.h"
#include "nc.h"

int nc_exit_srv;
int nc_ex_spawned = 0;
#ifdef OS9
int nc_os9fork = 0;
#endif
extern unsigned long nc_select_mask;
int nc_rx_fd = -1;      /* to allow multiply protocol select */

/*
 * This procedure loops forever processing rpc's. It prints error messages
 * to the stdout when verbose is non-zero. The exit condition specifies
 * when exit(status) or return(status) should be called. Useful values:
 * NC_EX_ONFORK will cause exit after the first call if this is the spawned server.
 * NC_RET_NOW will cause return after one procedure has been processed.
 * Timeout is used with the exit value NC_EX_TIMEOUT and will cause exit
 * if expired without getting a new call.
 * The global value nc_exit_srv can be set by the user procedure to change
 * the exit condition. Exit values NC_EX_ONFORK, NC_EX_NOW and NC_EX_NEXT
 * are in effect only for the spawned process.
 * If p is equal to NC_P_FIRST, the protocol of the first exported service
 * will be used.
 */
int
nc_serverloop(long unsigned int timeout, protocol *p, \
              int whenexit, int verbose)
/* timeout in milliseconds */
            
                      
{
	register nct_cservice *cd;
	int sid, status;
	int do_select;
	struct timeval tmo;
	fd_set mask;
	protocol *prot;
  
	if (p == NC_P_FIRST) {
		/* Check if any services are exported at all */
		if( (sid = nc_sfirst( NC_EXPORTED )) < 0 ) {
			return(-1);
		}

		/* Get the current protocol to be used. For now just */
		/* use the protocol given in the first service. */
		if( (cd = nc_sgetptr( sid )) == 0 ) {
			RETURNERR( ENCNOSERVICE, ETNCRT, "nc_sgetptr()" )
		}
		if ((p = cd->ncc_protocol) == 0) {
			RETURNERR(ENCCNOPROT, ETNCRT, "")
		}
	}


	if (whenexit & NC_EX_RWAIT) {
		/* rwait compatibility */
		if (!( timeout = cd->ncc_noresponse))
			/* 0 means very long time */
			timeout = 0x7fffffff;
		whenexit = NC_EX_TIMEOUT;
	}

	/*
	 * Set the global value to what the user specified so that it
	 * can later be checked if we fork.
	 */
	nc_exit_srv = whenexit;

	/*
	 * Verify whether we should do a select on rx and other protocols
	 * simultaneously or because timeout has been specified.
	 * Will work only for some protocols (tcp) for the time being.
	 */
	if ((p != NC_P_RX) && ((nc_rx_fd != -1) || (whenexit & NC_EX_TIMEOUT)))
		do_select = 1;
	else
		do_select = 0;
	prot = p;
	/*
	 * Loop forever processing calls. Print an error if any.
	 * Check exit condition to see whether we should exit.
	 */
	for( ;; ) {
#if defined(PROT_SELECT)
		if (do_select) {
			tmo.tv_sec = timeout/1000;
			tmo.tv_usec = timeout%1000;
			mask = nc_select_mask;
			if(select(&mask, 0, 0, &tmo) < 0)
				RETURNERR(errno, ETUNIX, "select()");
			if ((nc_rx_fd != -1) && (mask & nc_rx_fd))
				prot = NC_P_RX;
			else
				/* for the time being we assume tcp if it was not rx */
				prot = p;
		}
#endif
		if (((status = nc_getcall(prot, timeout)) < 0) && verbose) {
#ifdef OS9			/* fprintf(stderr, "server %s :", cd->ncc_name);  LP - 23.08.2000 */
                     fprintf(stderr, "%s: ", gettime());
#else
                     time_t locatime=time(0);  /* CS - 05.09.2000 */
		     fprintf(stderr,"%s :",ctime(&locatime));
#endif
		     nc_perror("nc_getcall()");
		} 
    /*  LP - 23.08.2000 - debugging *
      else {
      fprintf(stderr, "serverloop: \n");
      fprintf(stderr, "  cd->ncc_link = 0x%x\n  cd->ncc_name = 0x%x\n", \
              cd->ncc_link, cd->ncc_name);
    }
    * ---------------------------------- */
#ifdef OS9
		/* for OS9 we have to check if we should fork the new master */
		if (nc_os9fork)
			if (prot == NC_P_RX) {
				/*
				 * In OS9 the rx_fork() swaps the port and fd
				 * used by the rx to the new one and spawns the
				 * new instance of the server.
				 */
				nc_os9fork = 0;
				rx_fork();
		}
#endif
		TRACE(calls && fprintf(stderr, "serverloop: tmo %ld nc_exit_srv %x nc_ex_spawned %d select %d status %d\n",
			timeout, nc_exit_srv, nc_ex_spawned, do_select, status));
		whenexit = nc_exit_srv;
		if (nc_ex_spawned) {
			/* check whether spawned process should exit */
			switch (whenexit & NC_EX_SPAWNED) {
			    case NC_EX_ONFORK  /* = NC_EX_NOW */:
#if defined(_NO_ATEXIT)
				nc_freeall();
#endif
				exit(0);

			    case NC_EX_EXTNXT:
				nc_exit_srv = ((whenexit & NC_EX_PARENT) | NC_EX_NOW);
				break;
			    case NC_EX_TIMEOUT:
				if (status == 0) {
#if defined(_NO_ATEXIT)
					nc_freeall();
#endif
					exit(errno);
				}
			    case NC_EX_CUSTOM:
				if (status == 0)
					return(status);
			    default:
				break;
			}
		} else {
			/* check whether we really should exit */
			switch (whenexit & NC_EX_PARENT) {
			    case NC_EX_RETNXT:
				nc_exit_srv = ((whenexit & NC_EX_SPAWNED) | NC_EX_RETNOW);
				break;
			    case NC_EX_RETNOW:
				return(status);
			    case NC_EX_RETTMO:
				if (status == 0)
					return(status);
				break;
			    default:
				break;
			}
		}
	}
}


/*
 * This call is provided to wait on multiply protocols and to select on a user
 * mask. The system select will be performed on the user mask or'ed with the mask
 * used by nc services. It returns number of descriptors with pending events,
 * the mask of the pending events found, and the protocol in case that an
 * event was found on the nc mask.
 */
#if  defined(PROT_SELECT)
int
nc_sselect(long timeout, long* mask,protocol ** p)
{
	struct timeval tmo;
	long fullmask;
	int status;

	tmo.tv_sec = timeout/1000;
	tmo.tv_usec = timeout%1000;
	fullmask = *mask | nc_select_mask;
	if((status = select(&fullmask, 0, 0, &tmo)) < 0)
		RETURNERR(errno, ETUNIX, "select()");
	if (fullmask & nc_rx_fd)
		*p = NC_P_RX;
	else
		/* for the time being we assume the p if it was not rx */
		;
	return(status);
}
#endif

/*
 * The old interface retained for compatibility reasons. Remove it sometime.
 */

int
rwait(int policy, int maxentries)
{
	return(nc_serverloop(0xffffffL, NC_P_FIRST, NC_EX_RWAIT, 1));
}
