/* rx - remote execution protocol. The common part.
	The protocol can be entirely in user space (RX_USER) or in kernel
	space (RX_KERNEL_K) with the interface in user space (RX_KERNEL_U).
	The kernel implementation is for the Xenix TCP/IP driver

	Creation:
		August 1989 K. Kostro, CERN/SPS
		Oct. 1989       user code implementation using socket interface
		Dec. 1989       limited kernel implementation to test performance
		Feb. 1990       final implementation in user & kernel code
		May. 1990       Included in the NC/RPC generation

  LP - 23.05.2000 - include files fixes
                  - correct cast in bind() calls
                  - argument len in handle_rx_pkt() becomes int from u_int
                  - correct cast in sendto() and recvfrom() calls
 * CS - 07.09.2000 : general heavy clenaup
 *                   removed all old dead code: RX_KERNEL etc...		  

*/
#include "allincl.h"
#ifndef OS9
  #include <sys/time.h> /*CS sept 2000*/
#endif 


/*
 * Global variables for the rx protocol.
 */
struct rx_glob rx_glob = {1L, RX_INIT_WINDOW, 0L, 0L, {0}, {0}, 0,
	(struct inout_pkt *)0, {0}, 0L, 0, (struct inout_pkt *)0, {0}};

struct rx rx = {0};
struct rx_stats rx_stats = {0};
/*
 * If network_lossage is a number n > 0,
 * send_segment() will randomly drop 1 out of every n segments on the floor.
 */
int network_lossage = 0;


/*
 * Initialises the communication for rx i.e. makes an UDP port for all
 * rx communication if not already done. Returns the rx file descriptor
 * or (-1) and errno set in case of error.
 */

public int
rx_init(short unsigned int *port)
{
	struct sockaddr_in sin_from;
	int sinlen = sizeof(struct sockaddr_in);
	struct timeval no_activity;
	int status;

/*
   Initialise global variables and the local communication endpoint.
   rx is using a single UDP socket for all communication.
   If *port == 0, any free port will be allocated.
   Returns the file descriptor on success, -1 if fails (errno and errtype
   will be set). The port will be set to the allocated port.
   To allow forking a second socket descriptor can be established using
   the rx_newsd() call.
*/
	TRACE(calls && fprintf(stderr, "rx_init() \n"));
	if (rx.initialised != 0)
		if (*port == rx.port || *port == 0) {
			TRACE(assoc && fprintf(stderr, "udp port %d was already allocated\n", rx.port) );
			return(rx.sd);
		} else {
			RETURNERR(EINITIALIZ, ETRX, "already initialised");
		}
	if ((rx.sd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		RETURNERR(errno, ETUNIX, "socket()");

#if !defined(NOMEMLIB)
	memset((char*)&sin_from, '\0', sizeof(struct sockaddr_in));
#else
	/* another brand of memory copy functions from BSD */
	bzero((char*)&sin_from, sizeof(struct sockaddr_in));
#endif
	sin_from.sin_family = AF_INET;
	sin_from.sin_addr.s_addr = htonl(INADDR_ANY);
	sin_from.sin_port = htons(rx.port = *port);
	if ((status = bind(rx.sd, (struct sockaddr *)&sin_from, \
                     sizeof(struct sockaddr_in))) < 0)
		RETURNERR(errno, ETUNIX, "bind()");


	/*
	 * The rx protocol implementation in user space is using the following
	 * strategies to minimise the overhead and/or signal interference:
	 * Fast mode:
	 *      In this mode packets will not be acknowledged if not necessary
	 *      This mode should typically be used if the client/server
	 *      communication is frequent and intervalls between calls are
	 *      short (less than 5 sec). Interrupts can be enabled or
	 *      disabled.
	 * ack mode (conservative mode):
	 *      In this mode we try to reduce interrupts by explicitely
	 *      acknowledging calls/replies. In user space implementations
	 *      we explicitely ack the reply in the client and the call in
	 *      the server to avoid receiving retransmissions. To avoid being
	 *      blocked until the next call the server also explicitely asks
	 *      client to acknowledge the reply. Interrupts remain enabled in
	 *      the server and for asynch client calls, so that we do not miss
	 *      any call or reply. This is default mode.
	 * nosignal mode:
	 *      Similar to the secure mode but signals are always disabled.
	 *      Servers can miss a new request when the processing of the
	 *      previous one has not been yet finished. The same can happen
	 *      to asynchronous replies. This mode can be securely used
	 *      when the time between receiving calls (rx_get_call(),
	 *      rx_get_reply()) is shorter than the call/reply timeout
	 *      (currently 3 RX_MS_TICK's) or with servers which serve a single
	 *      client (for instance fork to handle new requests).
	 *      this is default mode.
	 * signal mode:
	 *      In this mode we turn on net signals or polling every time
	 *      we go back to the user code. This prevents from loosing
	 *      packets but we take risk of interrupted system calls. This
	 *      mode is switched on with RX_SIGON.
	 *
	 *      Although any combination of intr mode and ack mode can be
	 *      used, we concentrate on nointr and ack mode for secure
	 *      transparent communication paying some price of performance
	 *      loss and of functionality loss in special cases.
	 *      Other modes can be enabled in cases which are particullary
	 *      adapted.
	 *
	 *      Retransmissions can be disabled for unreliable communication
	 *      by setting the RX_NORETR flag.
	 */

	/* Set up the interrupt for UDP packets or alarms if the previous is
	   not available */

#ifdef NET_ASYNCH
#ifdef BSDnetsig
	{
	int pid;
	pid = getpid();
	if (fcntl(rx.sd, F_SETOWN, -pid) != 0)
		RETURNERR(errno, ETUNIX, "fcntl(,F_SETOWN,)");
	if (fcntl(rx.sd, F_SETFL, FASYNC) != 0)
		RETURNERR(errno, ETUNIX, "fcntl(,F_SETFL,)");
	}

	sv.sv_handler = net_intr;
	sv.sv_mask = INTERRUPT_MASK;
	sv.sv_onstack = 0;
	if (sigvec(SIGIO, &sv, 0) != 0)
		RETURNERR(errno, ETUNIX, "sigvec()");

	{
		struct itimerval it;
		struct sigvec sv;
		sv.sv_handler = rx_tick;
		sv.sv_mask = INTERRUPT_MASK;
		sv.sv_onstack = 0;
		if (sigvec(SIGALRM, &sv, 0) != 0)
			RETURNERR(errno, ETUNIX, "sigvec(SIGALRM,)");
	}
/*
	it.it_value.tv_usec = 0L;
	it.it_value.tv_sec = RX_TICK;
	if (setitimer(ITIMER_REAL, &it, 0) != 0)
		RETURNERR(errno, ETUNIX, "setitimer()");
*/

	/* stop the signal delivery for the time being */
	x = sigblock(INTERRUPT_MASK);
#endif
#ifdef SYSVsignal
#ifdef SYSVstreamio
/* I did not found a system on which this would work so far !!!! */
	/*
	 * stream driver implementations of TCP/IP should deliver SIGPOLL
	 * on 'selectable' events pending.
	 */
	if (ioctl(rx.sd, I_SETSIG, (S_INPUT | S_HIPRI | S_MSG)) != 0)
		RETURNERR(errno, ETUNIX, "ioctl(,I_SETSIG,)");
	sigset(SIGPOLL, net_intr);
#endif
#endif
#endif /* NET_ASYNCH */

#ifdef NET_POLL
	/* set up the alarm handling but do not set alarm yet */
#ifdef SYSVtime
	sigset(SIGALRM, rx_tick);
#endif
#endif /* NET_POLL */

	if (rx_glob.current_packet == (struct inout_pkt *)0)
		rx_glob.current_packet = rx_alloc_packet();
	if (rx_glob.out_packet == (struct inout_pkt *)0)
		rx_glob.out_packet = rx_alloc_packet();

	tm_now(&no_activity);
	rx_glob.no_activity = SEC(&no_activity);

	rx.mode = RX_ACK | RX_NOSIG;
	rx.retransmissions = RETRANSMISSIONS;
	rx.retr_tmo =  RETRANSMIT_TMO;
	rx.freelists = 1;

#ifdef DEBUG
	/* set id and routine to enable tracing */
	settrace("rx", rx_trace);
#if !(defined(OS9))
	srandom(no_activity.tv_sec);         /* initialise random number generation */
#endif
#endif

	rx.initialised = 1;
	rx.newsd = -1;

	/* bind does not get the port so we have to get it explicitely */
	if (getsockname(rx.sd, (struct sockaddr *)&sin_from, &sinlen) < 0)
		RETURNERR(errno, ETUNIX, "getsockname()")
	*port = rx.port = ntohs(sin_from.sin_port);

	TRACE(calls && fprintf(stderr, "rx_init() returning %d\n", rx.sd));
	return(rx.sd);
}


public int rx_closeup(void)
{
	register struct association *a;

	TRACE(calls && fprintf(stderr, "rx_closeup() \n"));
	if (close(rx.sd) < 0)
		RETURNERR(errno, ETUNIX, "close()")
	/* traverse active associations cleaning up */
	a = (struct association *)RXHEAD;
	while ( (char *)a != NULL) {
		rx_cleanup_ass(a);
		/* list destroyed, start again */
		a = (struct association *)RXHEAD;
	}

	/* clean up packet free list */
	rx_garbage_collect();

	rx.initialised = 0;
	return(0);
}




static void
handle_rx_pkt(register struct rx_hdr *h, int len)  /* LP - was unsigned int */
                          
                /* length of rx data (without header) */
{
	register struct association *a;
	unsigned long call_number;
	unsigned short seq_number;
#ifdef DEBUG
	struct addr debug_addr;
#endif

	if (len < 0)
		return;

	/* find the association */

	TRACE(calls && fprintf(stderr, "handle_rx_pkt() \n"));
	/*
	 * 13.05.93 kk modified finding association because reply can
	 * have an different source address than the call destination !!
	 */
	call_number = ntohl(h->call_number);
/*
	for (a = (struct association *)RXHEAD; (char *)a != NULL &&
		(a->addr_peer.host != rx_glob.current_host ||
		    a->addr_peer.port != rx_glob.current_port ||
		    ((a->type == RX_SERVER) && (h->pk_type == RX_RETURN_DATA || h->pk_type == RX_CALL_ACK)) ||
		    ((a->type == RX_CLIENT) && (h->pk_type == RX_CALL_DATA || h->pk_type == RX_RETURN_ACK))
		);
		a = (struct association *)a->link)
	;
*/
	for (a = (struct association *)RXHEAD; (char *)a != NULL &&
		( a->addr_peer.port != rx_glob.current_port
		    || ((a->type == RX_SERVER) &&
			((a->addr_peer.host != rx_glob.current_host ||
			h->pk_type == RX_RETURN_DATA ||
			h->pk_type == RX_CALL_ACK)))
		    || ((a->type == RX_CLIENT) &&
			((a->call_number != call_number) ||
			(h->pk_type == RX_CALL_DATA ||
			h->pk_type == RX_RETURN_ACK)))
		);
		a = (struct association *)a->link);
	if ((char *) a == NULL)
		switch (h->pk_type) {
		    case RX_CALL_DATA :
			if (h->seq_number != 0) {
				rx_stats.dr_ass ++;
				return;
			}
			/* this is a new call */

			a = rx_alloc_assoc();
			a->addr_peer.host = rx_glob.current_host;
			a->addr_peer.port = rx_glob.current_port;

			TRACE(assoc && fprintf(stderr, "allocated new assoc at %lx for addr %s\n",
				(long)a, print_ipaddr(&a->addr_peer)));

			a->call_number = 0L;
			a->type = RX_SERVER;
			a->ourwindow = rx_glob.window;
			break;
		    case RX_RETURN_DATA:
			/* send an error packet */
			if (!(h->flags & NO_PEER)) /* do not send it forever */
			  send_err((unsigned long)rx_glob.current_host,
			    rx_glob.current_port, 0, RX_RETURN_DATA, NO_PEER);
		    default:
			/* drop it */

#ifdef DEBUG
			debug_addr.host = rx_glob.current_from.sin_addr.s_addr;
			debug_addr.port = rx_glob.current_from.sin_port;
#endif
			TRACE(packets && fprintf(stderr,
			    "no association for pkt len %d type %d c_n %d seq %d ww %d flg %hx from %s\n",
			    (int)len, (int)h->pk_type, (int)call_number, (int)ntohs(h->seq_number), (int)h->window, h->flags,
			    print_ipaddr(&debug_addr)));
			rx_stats.dr_ass ++;
			return;
		} /* switch */
	else {
		TRACE(assoc && fprintf(stderr, "found assoc at %lx for addr %s\n",
			(long)a, print_ipaddr(&a->addr_peer)));
		if ((h->flags & NO_PEER) && (h->pk_type != RX_CALL_DATA)) {
			/* NO_PEER AND UNREL_FLG have the same values */
			a->state = ERROR_STATE;
			if (h->pk_type == RX_CNTRL)
				/* this is coming from the ip dispatcher */
				a->error = ENOADDR;
			else
				a->error = ENOPEER;
			a->tm = 0;
			a->w_intr = 0; /* release wait for interrupt */
		}

	}
	/*
	 * Now we found the association so we can handle the packet.
	 */
	a->hiswindow = h->window;
	seq_number = ntohs(h->seq_number);

	TRACE(packets && fprintf(stderr,
		"received pkt len %d type %d c_n %ld seq %d ww %d flg %x\n",
		len, h->pk_type, call_number, (int)seq_number, h->window, h->flags));

	switch (h->pk_type) {
	    case RX_CALL_DATA :
		/*
		 * Data packet from client. The normal case is a new call
		 * or a segment of the current call. If we are receiving
		 * or sending this might be a retransmitted packet.
		 */
		rx_stats.cd_rcv ++;
		ASSERT(a->type == RX_SERVER)
		/*
		 * We accept both types of return acknowlegement: explicit
		 * ACK and a new call.
		 */
 
		if ((seq_number == 0) && ((call_number > a->call_number)||call_number==1) &&   /*C.S. 28 dec 2000-new call can be on same old host:port and handled in old assoc*/
			    (a->state == SENDING || a->state == SENT || a->state == RECEIVED)) {
			/* the first packet of a new call acks our return */

			/* we do not accept new calls until the previous one has
			   been handled */
			if ((a->in_pkts.head != NULL) && (a->state == RECEIVED)) {
				/*
				 * Kernel implementation does not discard old calls
				 * We would need special state or flags to protect
				 * against simultaneous reading by get_call()
				 */

				/* if no-reply call, send BACK_OFF , otherwise discard it */
				if (a->prot_flags & NOREPL_FLG) {
					/* The client does not know that we did
					   not process the previous call. That's normal
					*/
					send_cntr(NIP, a, RX_CALL_ACK, BACK_OFF);
					return;
				} else {
					/* New call cancels the old one */
					 while (a->in_pkts.head != (struct queue_item *)0)
						rx_free_packet(q_rhead(&a->in_pkts));
					/* fall through to set the new call */
				}

			}
			a->call_number = call_number;
			a->hdr.call_number = htonl(a->call_number);
			a->state = RECEIVING;
			a->tm = 0;
			a->seq = 0;
			/* we timestamp the first packet so that we can clean up partial calls */
			tm_now(&a->timestamp);
			/* we would free here any packets saved for retransmission */
		} else
			if ((seq_number == a->seq + 1) &&
				    (call_number == a->call_number) &&
				    (a->state == RECEIVING)) {
				/* subsequent segment of the current call */
				a->seq ++;
			} else {
				if ((call_number == a->call_number) &&
					    (a->state == RECEIVING ||
					    a->state == RECEIVED)) {
					/* retransmitted packet, ack the last received */
					send_cntr(NIP, a, RX_CALL_ACK, ACK);
					rx_stats.retr_rcv ++;
					return;
				}
				if ((call_number == a->call_number) &&
					    (a->state == SENDING)) {
					/* unusual case of retransmitted packet.
					   Since we are already sending the
					   reply, retransmit packet */

					TRACE(exceptions && fprintf(stderr, "retransmitted packet call no %d current call %d\n",
						call_number, a->call_number) );

					if (a->w_intr != 0) {
						a->hiswindow = 1;
						a->w_intr = 0; /* release wait for interrupt */

					}
					rx_stats.retr_rcv ++;
					return;
				}
				/* we do not expect this packet */
				rx_stats.dr_callno ++;
				/* 17.02.94 kk Do not cleanup association in SENT state
				   Otherwise no-reply messages can be received twice if
				   the previous ack has been lost
				*/

				TRACE(exceptions &&
				fprintf(stderr, "not expected call pkt %d in state %d current call %d\n",
					call_number, a->state, a->call_number));
				if (a->state != SENT)
					rx_cleanup_ass(a);

				return;
			}
		/* handle and queue call packet */

		q_atail(&a->in_pkts, (struct queue_item *)rx_glob.current_packet);
		rx_glob.current_packet->len = len;
		rx_glob.current_packet = (struct inout_pkt *)0;


		/* check if this is the final segment */
		if (h->flags & PLEASE_ACK) {

			send_cntr(NIP, a, RX_CALL_ACK, ACK);
			a->w_intr = 0; /* release wait for interrupt */

		}

		if (h->flags & LAST_SEGMENT) {
			if (h->flags & NOREPL_FLG)
				a->prot_flags |= NOREPL_FLG;
			if (h->flags & UNREL_FLG)
				a->state = SENT;
			else
				a->state = RECEIVED;
			a->tm = 0;

			TRACE (timers && fprintf(stderr, "set timer to 0 ass %lx\n",
				(long)a) );
			tm_now(&a->timestamp);

			a->w_intr = 0; /* release wait for interrupt */
			/* ack if not already done (unless fast mode or broadcast) */
			if (!(h->flags & PLEASE_ACK) && !(h->flags & UNREL_FLG) &&
				    (rx.mode & RX_ACK))
				send_cntr(NIP, a, RX_CALL_ACK, ACK);

		}
		return;

	    case RX_RETURN_DATA:
		/* we expect only returns for the current call or retransmissions */
		rx_stats.rd_rcv ++;
		ASSERT(a->type == RX_CLIENT)
		if ((seq_number == 0) && (call_number == a->call_number) &&
			    (a->state == SENDING || a->state == SENT)) {
			/* the first return packet acks our call */
			a->state = RECEIVING;
			a->seq = 0;
			/* we would free here any packets saved for retransmission */
		} else {
			if ((seq_number == a->seq + 1) &&
			    (call_number == a->call_number) &&
			    (a->state == RECEIVING || a->state == RECEIVED))
			/* subsequent segment of the current return */
			a->seq ++;
			else {
				if ((call_number == a->call_number) &&
					    (a->state == RECEIVING ||
					     a->state == RECEIVED)) {
					/* retransmitted packet, ack the last received */
					send_cntr(NIP, a, RX_RETURN_ACK, ACK);
					rx_stats.retr_rcv ++;
					return;
				}
				if ((call_number < a->call_number) &&
					    (a->state == SENDING)) {
					/* unusual case of retransmitted packet.
					   Since we are already sending the
					   new call, retransmit packet */

					TRACE(exceptions && fprintf(stderr, "retransmitted packet reply no %d current call %d\n",
						call_number, a->call_number) );

					if (a->w_intr != 0) {
						a->hiswindow = 1;
						a->w_intr = 0; /* release wait for interrupt */
					}
					rx_stats.retr_rcv ++;
					return;
				}
				/* we do not expect return data */
				rx_stats.dr_typ ++;

				TRACE(exceptions &&
				fprintf(stderr, "unexpected return pkt %d in state %d current call %d\n",
					call_number, (int)a->state, (int)a->call_number));

				return;
			}
		}
		/* handle and queue return packet */
		/*
		 * In some cases such as blocking rx_call the space for the
		 * return data has been already allocated. In this case we
		 * can avoid copying twice by copying directly to the user
		 * buffer (only when protocol in user space).
		 */

		{ int prevlen;
		    if (((prevlen = a->seq * RX_SIZE) + len) > a->rlen) {
			/* receive buffer too short, queue the packet */
			q_atail(&a->in_pkts, (struct queue_item *)rx_glob.current_packet);
			if ((prevlen = a->rlen - prevlen) > 0) {
				/* copy up to the buffer size */
				bcopy((char *)h + sizeof(struct rx_hdr), a->rmsg + a->seq * RX_SIZE, prevlen);
				rx_glob.current_packet->len = len - prevlen;
				a->readp = (char *)h + sizeof(struct rx_hdr) + prevlen;
				/*
				 * we could also release the client's wait here instead waiting
				 * for the complete message
				 */
			} else {
				rx_glob.current_packet->len = len;
			}
			rx_glob.current_packet = (struct inout_pkt *)0;
		    } else
			bcopy((char *)h + sizeof(struct rx_hdr), a->rmsg + a->seq * RX_SIZE, len);
		}

		/* check if this is the final segment */
		if (h->flags & PLEASE_ACK) {

			send_cntr(NIP, a, RX_RETURN_ACK, ACK);
			a->w_intr = 0; /* release wait for interrupt */

		}
		if (h->flags & LAST_SEGMENT) {
			a->state = RECEIVED;
			a->tm = 0;

			TRACE (timers && fprintf(stderr, "set timer to 0 ass %lx\n",
				(long)a) );

			a->w_intr = 0; /* release wait for interrupt */


			/* ack if not already done (unless fast mode) */
			if (!(h->flags & PLEASE_ACK) && (rx.mode & RX_ACK))
				send_cntr(NIP, a, RX_RETURN_ACK, ACK);
		}
		return;


	    case RX_CALL_ACK:
		rx_stats.ca_rcv ++;
		if ((h->flags & BACK_OFF) && (a->state == SENDING) ){
			if(a->tm_modifier < RETRANSMISSIONS + 2) {
				a->tm_modifier = RETRANSMISSIONS + 2;
				a->tm = BACKOFF_RETR_TMO;
			}
			/* otherwise we switched already to the
				retransmission mode */
			return;
		    }
		rx_stats.ra_rcv --;     /* because we dont break */
	    case RX_RETURN_ACK:
		rx_stats.ra_rcv ++;
		if (call_number != a->call_number) {
			rx_stats.dr_callno ++;

			TRACE(exceptions &&
			fprintf(stderr, "not expected ack pkt call %d in state %d current call %d\n",
				call_number, a->state, a->call_number));
			return;
		}
		if (a->state == SENDING)
		    if (seq_number >= a->seq) {
			/* intermediate ack in case of multiple segments or timeout */
			a->seq = seq_number + 1;
			ASSERT(h->window < MAXWINDOW)
			a->hiswindow = h->window;
			if (seq_number == a->last_seq) /* acks the whole message */
				a->state = SENT;
			a->tm = 0;

			TRACE (timers && fprintf(stderr, "set timer to 0 ass %lx\n",
				(long)a) );

			a->w_intr = 0; /* release wait for interrupt */
		    } else
			if (seq_number == a->seq - 1) {
				ASSERT(h->window < MAXWINDOW)
				a->hiswindow = h->window;
			} else {
				/* we have seen it already or it's invalid */
				rx_stats.dr_typ ++;
				return;
			}
		return;
	    case RX_CNTRL:
		/* used for probe and probe ack, to abort association */
		if(h->flags & ASS_ABORT) {
			rx_cleanup_ass(a);
			return;
		}
		if(h->flags & PLEASE_ACK) {
			/* it's a probe, reply with ACK */
			send_cntr(NIP, a, RX_CNTRL, ACK);
			return;
		}
		if(h->flags & ACK) {
			/* probe reply, adjust timeouts */
#if defined(PROBE_SERVER)
/* for the time being ignore probe reply. */
			if (a->tm_modifier < NO_REPLY_TMO)
				a->tm_modifier = NO_REPLY_TMO;
			a->tm = a->tm_modifier;
			a->tm_action = no_reply;
#endif

			TRACE (timers && fprintf(stderr, "probe reply: set timer to %ld modifier to %d ass %lx",
				a->tm, a->tm_modifier, (long)a) );
			return;
		}
	    default:
		/* drop it */
		rx_stats.dr_typ ++;
		return;
	} /* switch */
}


void
do_retransmit(struct association *a)               
{
	/*
	 * This routine will be called asynchronously after retransmission timeout.
	 * The actual retransmission will be done by the sending task which
	 * is waiting to resume.
	 * The purpose of the first retransmission is mainly to cope with lost
	 * packet while the second retransmission timeout is longer to give
	 * servers which use user code implementations the chance to be swapped
	 * in and the packet handling routine to be called.
	 */

	TRACE(calls && fprintf(stderr, "do_retransmit() \n"));

	rx_stats.retr_snd ++;
	/*
	 * Special treatment for back-off packets:
	 * BACK_OFF packet in SENDING state means that the receiving host
	 * run out of packet list, or the previous NOREPLY call has not
	 * yet been processed.
	 * We apply the back-off retransmission timeout up to 10 times.
	 */
	if ((a->tm_modifier ++) >= RETRANSMISSIONS + 2) {
		if ((a->tm_modifier) >= RETRANSMISSIONS + 10) {
			a->error = ETOSERV;
			a->state = ERROR_STATE;
		} else {
			a->hiswindow = 1;
			a->tm = BACKOFF_RETR_TMO;
		}
	} else {
		if (a->tm_modifier > rx.retransmissions) {
			a->error = ETOSERV;
			a->state = ERROR_STATE;
		} else {

			a->tm = rx.retr_tmo << a->tm_modifier;
			a->hiswindow = 1;
		}
	}

	TRACE (timers && fprintf(stderr, "retr timeout: set timer to %ld modifier to %d ass %lx",
		a->tm, a->tm_modifier, (long)a) );

	a->w_intr = 0;
}


#if defined(PROBE_SERVER)
void
probe_retr(struct association *a)

	/*
	 * This function gets called after probe timeout.
	 */
{

	TRACE(calls && fprintf(stderr, "probe_retr() \n"));

	/* in case of probe timeout we are back to retransmission timeout */
	if (a->tm_modifier > NO_REPLY_TMO) {
		a->tm_modifier = 0;
	} else {
		if ((a->tm_modifier ++) >= RETRANSMISSIONS) {
			a->error = ETOREPL;
			a->state = ERROR_STATE;
			a->w_intr = 0;
		}
	}
	a->tm = RETRANSMIT_TMO*2;
	if (send_cntr(NIP, a, RX_CNTRL, PLEASE_ACK) < 0) {
		a->error = errno;
		a->state = ERROR_STATE;
		a->w_intr = 0;
	}
}
#endif /* PROBE_SERVER */


void
no_reply(struct association *a)

                      
	/*
	 * This function gets called when there is no reply after the call
	 * has been ack'ed.
	 * The probe packet is sent in exponentially increasing time intervals
	 * given the server always replies. For probe retransmisssion the normal
	 * retransmission delays aply.
	 */
{
	struct timeval time_now;
	int left;

	TRACE(calls && fprintf(stderr, "no_reply(,%lx) \n",(long)a));

#if defined(PROBE_SERVER)
	/* in case of probe timeout we are back to retransmission timeout */
	if ((a->tm_modifier *=2) >= MAX_NO_REPLY_TMO)
		a->tm_modifier = MAX_NO_REPLY_TMO;
	a->tm = RETRANSMIT_TMO * 2;     /* first timeout slightly longer */
	a->tm_action = probe_retr;

	TRACE (timers && fprintf(stderr, "no repl timeout: set timer to %ld modifier to %d ass %lx",
		a->tm, a->tm_modifier, (long)a) );

	if (send_cntr(NIP, a, RX_CNTRL, PLEASE_ACK) < 0) {
		a->error = errno;
		a->state = ERROR_STATE;
		a->w_intr = 0;
	}
#else /* PROBE_SERVER */
	/* check if the timeout really expired */

	tm_now(&time_now);
	left = a->user_tmo - MS_LEFT(&time_now, &a->timestamp);
	if (left <= 0) {
		a->error = ETOREPL;
		a->state = ERROR_STATE;
		a->w_intr = 0;
	} else {
		a->tm = (left/RX_MS_TICK)*RX_MS_TICK;
		if (!a->tm)
			a->tm = RX_MS_TICK;
		a->tm_action = no_reply;
	}

#endif /* PROBE_SERVER */
}


static void
no_activity(long int now)
                        /* current time in ticks or seconds */
	/*
	 * This function gets called periodicaly to remove non-active associations
	 */
{
	register struct association *a;

	TRACE(calls && fprintf(stderr, "no_activity(%ld)\n", now));

	a = (struct association *)RXHEAD;
	while ( (char *)a != NULL) {
		if ((SEC(&a->timestamp) + NO_ACTIVITY_TMO < now) &&
			( (a->type == RX_CLIENT && (a->state == IDLE ||
			    a->state == ERROR_STATE || (a->state == RECEIVED &&
			    (char *)(struct inout_pkt *)a->in_pkts.head == NULL)
			))
			|| (a->type == RX_SERVER && (a->state == SENT || a->state == ERROR_STATE)) )
		) {
			 rx_cleanup_ass(a);
			/* list destroyed, start again */
			a = (struct association *)RXHEAD;
		} else
			a = (struct association *)a->link;
	}

	/* clean up free lists */
	if (rx_glob.net_wait_cnt > GARBAGE_COLLECT_CNT) {
		rx_glob.net_wait_cnt = 0;
		rx_garbage_collect();
	}
}


send_segment(struct association *a, char *buf, int len)
{
	struct sockaddr_in out_addr;

	TRACE(calls && fprintf(stderr, "send_segment() \n"));
	out_addr.sin_addr.s_addr = a->addr_peer.host;
	out_addr.sin_port = a->addr_peer.port;
	out_addr.sin_family = AF_INET;
	a->hdr.window = a->ourwindow;

	bcopy((char *) &a->hdr, rx_glob.out_packet->data, sizeof(struct rx_hdr));
	bcopy(buf, rx_glob.out_packet->data + sizeof(struct rx_hdr), len);

	TRACE(packets && fprintf(stderr, "sending data pkt len %d type %d c_n %d seq %d ww %d flg %x\n",
		len, a->hdr.pk_type, ntohl(a->hdr.call_number),
		ntohs(a->hdr.seq_number), a->hdr.window, a->hdr.flags));
	rx_stats.d_snd ++;
#ifdef DEBUG
#if !(defined(OS9))
	/*
	 * Randomly drop segments to exercise the protocol.
	 */
	if (network_lossage != 0 && (random() % network_lossage) == 0) {
		TRACE(exceptions && fprintf(stderr, "dropping data pkt\n"));
		return(0);
	}
#endif
#endif
	if (sendto(rx.sd, rx_glob.out_packet->data, len + sizeof(struct rx_hdr), 0, \
			(struct sockaddr *)&out_addr, sizeof(struct sockaddr_in)) < 0)
		RETURNERR(errno, ETUNIX, "sendto()");
	return(0);
}


/* called as macro send_cntr() */
send_cntrl(struct association *a, unsigned char type, unsigned char flags)
{
	/* send control packet */

	struct rx_hdr h;
	struct sockaddr_in out_addr;

	TRACE(calls && fprintf(stderr, "send_cntrl() \n"));
	out_addr.sin_addr.s_addr = a->addr_peer.host;
	out_addr.sin_port = a->addr_peer.port;
	out_addr.sin_family = AF_INET;

	h.call_number = a->hdr.call_number;
	h.seq_number = htons(a->seq);
	h.flags = flags;
	h.pk_type = type;
	h.window = a->ourwindow;
	rx_stats.c_snd ++;

	TRACE(packets && fprintf(stderr, "sending cntrl pkt type %d c_n %d seq %d ww %d flg %x\n",
		h.pk_type, ntohl(h.call_number),
		ntohs(h.seq_number), h.window, h.flags));
#ifdef DEBUG
#if !(defined(OS9))
	/*
	 * Randomly drop segments to exercise the protocol.
	 */
	if (network_lossage != 0 && (random() % network_lossage) == 0) {
		TRACE(exceptions && fprintf(stderr, "dropping cntrl pkt\n"));
		return(0);
	}
#endif
#endif
	if (sendto(rx.sd, (char *)&h, sizeof(struct rx_hdr), 0, \
			(struct sockaddr *)&out_addr, sizeof(struct sockaddr_in)) < 0)
		RETURNERR(errno, ETUNIX, "sendto()");
	return(0);
}


send_err(long unsigned int host, short unsigned int port, short unsigned int lport, unsigned char type, unsigned char flags)
{
	/*
	   Send error packet to host/port
	   Used to send error packet when we did not found matching
	   address or association. This will assure immediate notification
	   of peer.
	*/
	struct rx_hdr h;
	struct sockaddr_in     out_addr;

	TRACE(calls && fprintf(stderr, "send_cntrl() \n"));
	out_addr.sin_addr.s_addr = host;
	out_addr.sin_port = port;
	out_addr.sin_family = AF_INET;

	h.flags = flags;
	h.pk_type = type;
	rx_stats.c_snd ++;

	TRACE(packets && fprintf(stderr, "sending error pkt type %d c_n %d seq %d ww %d flg %x\n",
		h.pk_type, ntohl(h.call_number),
		ntohs(h.seq_number), h.window, h.flags));
#ifdef DEBUG
#if !(defined(OS9))
	/*
	 * Randomly drop segments to exercise the protocol.
	 */
	if (network_lossage != 0 && (random() % network_lossage) == 0) {
		TRACE(exceptions && fprintf(stderr, "dropping cntrl pkt\n"));
		return(0);
	}
#endif
#endif
	if (sendto(rx.sd, (char *)&h, sizeof(struct rx_hdr), 0, \
			(struct sockaddr *)&out_addr, sizeof(struct sockaddr_in)) < 0)
		RETURNERR(errno, ETUNIX, "sendto()");
	return(0);
}


/*      POLICY FOR SIGNALS IN RX

	This discussion reflects a long experience of asynchronous dealing
	with packet reception and time events in Unix. When a protocol
	such as RPC is implemented in user space one has to deal with
	asynchronous events of packet arrival and timeouts. If these events
	are implemented as signals, they can lead to system calls being
	interrupted which rends the whole package unusable. Of course one
	can attempt to block signals in crucial parts of code but there
	is still no guarantee that this will work in all cases and the package
	would become non transparent. The only really satisfactory solution
	is to put the protocol treatment into the kernel driver. Since this
	is not always possible a compromise has to be found. In our case
	the compromise is as follows:

	First of all we try to avoid getting interrupts. We acknowledge
	messages explicitely such avoiding possible retransmissions by
	the peer. We also implement timeouts, which serve administrative
	purposes such as cleaning up of structures, with timestamps rather
	then with timeouts and therefore signals. When we wait for something
	(e.g. return of a call) then we use select and check timeout queue
	at the same time. Applying these simple methods we should not need
	signal-driven events in the following cases:
	- in clients doing blocking calls, because no events are expected
	  between calls.
	- in clients doing asynch calls 'in groups' to get advantage of
	  paralel call execution, because they will wait processing returns
	  until all are done.
	- in servers which can finish processing of a call before another
	  call could time-out (which is order of 5 sec).
	- in servers which deal with only one client (i.e. fork'ed).

	To deal with all other cases (i.e. clients which do asynch calls
	to be able to do some processing (of undefined time) and servers
	which take a long time to serve a call but are not willing to fork)
	we (will) provide an interrupt mode to which we switch after a non
	blocking call in the client and after reception of call in the
	server. In this case signals cannot be turn off and overhead must
	be expected due to the switch of treatement. (couple of ms for every
	call).
*/


#ifdef NET_ASYNCH
static void
rx_tick()
{
struct association *a;
#ifdef NET_POLL
	rx_net_check();
#endif
	/*
	 * Traverse the association list in search for a timeout. The timer will
	 * be decremented unless equal to zero. If this is the last decrement
	 * the timeout action will be called.
	 */
	TRACE(timers && fprintf(stderr, "rx_tick() \n"));
	for (a = (struct association *)RXHEAD; (char *)a != NULL; a = (struct association *)a->link)
		if (a->tm) {
			a->tm --;
			if (!(a->tm)) {
				TRACE(timers && fprintf(stderr, "calling tm_action tm = %d modifier = %d\n",
				    a->tm, a->tm_modifier));
				(*(a->tm_action))(a);
			}
		}
	alarm(RX_TICK);
/*
fprintf(stderr, "rx_tick() returning \n");
*/
}
#endif


int
rx_net_wait(void)
{
/*
 * This procedure waits for packets simultaneously checking time events.
 */
	struct association *a;
	static struct timeval tick_time;
	int n;
	long nextwait;
	long mstowait = RX_MS_TICK;

	fd_set mask;

	int fromlen;
	struct timeval time_now, time_after;

	tm_now(&time_now);
	if (((rx_glob.net_wait_cnt++) % NET_WAIT_CNT) == 0) {
		if (SEC(&time_now ) > rx_glob.no_activity + NO_ACTIVITY_TMO) {
			rx_glob.no_activity = SEC(&time_now);
			no_activity(rx_glob.no_activity);
		}
	}

	for (;;) {
                FD_ZERO(&mask);
		FD_SET(rx.sd, &mask);
		tick_time.tv_sec = mstowait/1000;
		tick_time.tv_usec = (mstowait%1000)*1000;

		TRACE(timers && fprintf(stderr, "calling select tick time sec %ld usec %ld\n",
		    tick_time.tv_sec, tick_time.tv_usec) );

		/*if ((n = select(rx.sd + 1, (u_int32 *)&mask, 0, 0, &tick_time)) > 0){ *//*CS sept 2000*/
		if ((n = select(rx.sd + 1, &mask, 0, 0, &tick_time)) > 0) {
			TRACE(timers && fprintf(stderr, "rx_net_wait(): select returned %d\n", n));

			fromlen = sizeof(struct sockaddr_in);
			if ((rx_glob.current_length = \
           recvfrom(rx.sd, rx_glob.current_packet->data, \
			              RX_SIZE + sizeof(struct rx_hdr), 0, \
                    (struct sockaddr *)&rx_glob.current_from, &fromlen)) > 0)
			handle_rx_pkt((struct rx_hdr *)rx_glob.current_packet->data, rx_glob.current_length - sizeof(struct rx_hdr));
			if (rx_glob.current_packet == (struct inout_pkt *)0)
				rx_glob.current_packet = rx_alloc_packet();
			if (rx_glob.current_length < 0)
				RETURNERR(errno, ETUNIX, "recvfrom()");
			return(0);
		}

/*
fprintf(stderr, "select returned %d glob len %d\n", n, rx_glob.current_length);
*/
		/*
		 * Traverse the association list in search for a timeout. The timer will
		 * be decremented unless equal to zero. If this is the last decrement
		 * the timeout action will be called.
		 */
		if ((n < 0) && (errno == EINTR)) {
			/* we do not know where we were interrupted */
			tm_now(&time_after);
			mstowait = MS_LEFT(&time_after, &time_now);
/*
fprintf(stderr, "mstowait %d after %d.%d now %d.%02d\n", 
mstowait, time_after.tv_sec,time_after.tv_usec, time_now.tv_sec,time_now.tv_usec);
*/
			tm_now(&time_now);
		}
		if (RX.call_tmo) {
			RX.call_tmo -= mstowait;
			if (!(RX.call_tmo))
				/* timeout on call such as rx_get_call() */
				return(0);
			nextwait = RX.call_tmo;
		} else {
			/* quite long loop timeout */
			nextwait = RX_SELECT_TMO;
		}

		for (a = (struct association *)RXHEAD; (char *)a != NULL; a = (struct association *)a->link)
			if (a->tm) {
/*
fprintf(stderr, "tmo is %d\n", a->tm);
*/
				a->tm -= mstowait;
				nextwait = MIN(nextwait, a->tm);
				if (a->tm <= 0) {
					a->tm = 0;
					TRACE(timers && fprintf(stderr, "calling tm_action tm = %ld modifier = %ld\n",
					    a->tm, a->tm_modifier));
					(*(a->tm_action))(a);
					return(0);
				}
			}
		mstowait = nextwait;
	}
}


static int
rx_net_check(void)
{
	/*
	   Called by the user code signal mechanism on packet arrival.
	   Called also by polling routines (rx_wait() and rx_get_call()).
	   Can also be called by users to avoid missing packets.
	   Returns 0 if no packets found, 1 if packets found.
	*/
	struct association *a;
	static struct timeval immediately;
	int found;
	int n;

	fd_set mask;

	int fromlen;

	immediately.tv_sec = 0;
	immediately.tv_usec = 0;
	found = 0;

	do {
		FD_ZERO(&mask);
		FD_SET(rx.sd, &mask);
		TRACE(assoc && fprintf (stderr, "rx_net_check(): rx.sd is %d mask is %x\n",rx.sd, mask) );
		if ((n = select(rx.sd + 1, &mask, 0, 0, &immediately)) > 0) {
			TRACE(timers && fprintf(stderr, "rx_net_wait(): select returned %d\n", n));
			fromlen = sizeof(struct sockaddr_in);
			if ((rx_glob.current_length = \
                             recvfrom(rx.sd, rx_glob.current_packet->data, RX_SIZE + sizeof(struct rx_hdr), 0, \
                                      (struct sockaddr *)&rx_glob.current_from, &fromlen)) > 0)

			found = 0;
			handle_rx_pkt((struct rx_hdr *)rx_glob.current_packet->data, rx_glob.current_length - sizeof(struct rx_hdr));
			if (rx_glob.current_packet == (struct inout_pkt *)0)
				rx_glob.current_packet = rx_alloc_packet();
			if (rx_glob.current_length < 0)
				RETURNERR(errno, ETUNIX, "recvfrom()");
		}
	} while ( n > 0);
	if (found)
		return(1);
	else
		return(0);
}


rx_cleanup_ass(struct association *a)
{
	TRACE(assoc && fprintf(stderr, "cleanup ass %x\n", (long)a) );
	while (a->in_pkts.head != (struct queue_item *)0)
		rx_free_packet(q_rhead(&a->in_pkts));
	rx_free_assoc(a);
}


rx_garbage_collect(void)
{
	char *p;
	register struct association *a;

	/* force cleaning up without delay */
	a = (struct association *)RXHEAD;
	while ( (char *)a != NULL) {
		if ( (a->type == RX_CLIENT && (a->state == IDLE ||
			    a->state == ERROR_STATE || (a->state == RECEIVED &&
			    (char *)(struct inout_pkt *)a->in_pkts.head == NULL)
			))
			|| (a->type == RX_SERVER && a->state == SENT)
		) {
			rx_cleanup_ass(a);
			/* list destroyed, start again */
			a = (struct association *)RXHEAD;
		} else
			a = (struct association *)a->link;
	}

	while ((p = (char *)q_rhead(&rx_glob.free_packets)) != 0) {
		TRACE(allocation && fprintf(stderr,
			"rx_garbage_collect() free packet at %lx\n", (long)p));
		free(p);
	}

	return(0);
}



void
rx_dbg(void)
{
	struct association      *a;
	struct queue_item * pk;
	int cnt;

fprintf(stderr, "%8-s %5s %5s %5s %1s %1s %1s %2s %3s %3s %4s %1s\n",
      "assoc", "host", "port", "calln", "t", "s", "w", "er", "tm", "t-m",
      "seq", "q");
		for (a = (struct association *)RXHEAD; (char *)a != NULL;
			a = (struct association *)a->link) {
		/* count packet queue */
		for (pk = a->in_pkts.head, cnt = 0; pk != (struct queue_item *)0;
			pk = pk->link, cnt ++) ;

fprintf(stderr, "%8lx %2x.%2d %5d %5d %1d %1d %1d %2d %3d %3d %4d %1d\n",
	(long)a, ((a->addr_peer.host  >> 16) & 0xff),
	((a->addr_peer.host >> 24) & 0xff), a->addr_peer.port,
	a->call_number, a->type, a->state, a->w_intr, a->error,
	a->tm, a->tm_modifier, a->seq, cnt);
		}

}
