/* @(#)alarm.c	1.3.2.2 89/07/10 *
/*
 * Version 07/07/89
 * 14/04/89 raw: put error output under #ifdef DEBUG
 * 07/07/89 raw: add "alrmkill()" function
 *
 * alarm() function of the "alarm-package" for OS-9.
 * This package simulates the UNIX alarm() function for OS-9.
 * It consists out of
 * - this alarm() function;
 * - the alrmkill() function (see below);
 * - the alarmclock process.
 *
 * Usage: unsigned alarm(sec)
 *        unsigned sec;
 * Note: an executable "alarmclk" module must be present in the working
 *       executable directory.
 *
 * The user calls alarm(sec) from his program, as he would for the
 * UNIX routine. alarm() takes the input value and sends it over an
 * unnamed pipe to the alarmclock process; this process will in turn
 * send back a signal to the calling process after "sec" seconds.
 * A value of 0 disables the alarm clock.
 * As in the UNIX call, in repeated calls to alarm() the most recent
 * one overrides the previous ones. alarm() returns the amount of time
 * remaining from the previous call (i.e. 0 if the previous call was
 * not overridden and therefore the signal was sent); at the first call
 * from a program it always returns 0. In case of an error (see below)
 * FFFFFFFF is returned.
 * At the first call to alarm() from a program the alarm() function
 * opens an unnamed pipe and then forks the process alarmclk. In the
 * arguments of the fork call it transmits:
 * - its own process ID; this is needed to send the signal back;
 * - the file descriptor of the unnamed pipe.
 * Repeated calls will simply have their values written one after the
 * other into the pipe. The desired overriding will be done by the
 * alarmclk process; for details on this see the description of the
 * alarmclk process.
 * If the routine is interrupted and the interrupt handler calls itself
 * alarm(), the value that determines the setting of the alarm clock
 * is the value that was entered by the interrupting call. The value
 * returned by the interrupting call is the input value of the inter-
 * rupted call. The interrupted call will return the value it would
 * have returned without interruption. Repeated interruptions, even
 * interruptions of interruptions, will follow these same rules.
 *
 * If the alarmclk process is no longer needed, it can be terminated
 * with the alrmkill() function (see below). This should be done to
 * avoid clobbering the operating system with idle processes. 
 *
 * Errors during the execution of the alarm() function return in a 
 * return value FFFFFFFF. In addition, if DEBUG is enabled, a message
 * is printed to stderr.
 * These errors are:
 * - failure in forking the alarmclk process;
 * - pipe full (i.e., the alarmclk process probably died);
 * - other pipe write errors.
 *
 * Author: R. Wilcke (ESRF, Grenoble)
 */

/*
 * The following definition is neccessary for alarm() to work for 
 * the network compiler. It can be ignored or removed otherwise.
 */
#define OS9 OSK
#include <stdio.h>
#include <modes.h>
#include <errno.h>
#include <time.h>
#include <signal.h>

#define ALRMERR 0xffffffff
#define PIPDAT 4

extern char **environ;
extern int os9forkc();
static int alarmpid=-1;

unsigned alarm(sec)
unsigned sec;
{

   static unsigned char alrmlock=0;
   static int fildes,StdIn;
   static unsigned slpsec,secval,secsav,retval;
   static char ppid[6];
   static char *argblk[3]={"alarmclk",ppid,0};
   static time_t timeold;
   
   int status;
   unsigned timeleft;
   time_t timenow;

/*
 * Set up for possible interrruption. Store input value into global
 * variable immediately. Set interruption flag by atomic test-and-set
 * function. If flag was clear, continue processing. If flag was set,
 * an alarm() call is already processing: it will write the now up-
 * dated alarmclock value to the pipe; thus the interrupting call
 * just needs to define its return value and return.
 */
   secval=sec;
   retval=secsav;
   secsav=secval;
   if(tst_and_set(&alrmlock) != 0) return(retval);  

/* 
 * First call to alarm()?
 */
   if(alarmpid == -1)
   {
/*
 * Initialize time counters.
 */
      timeleft=0;
      time(&timeold);
/*
 * Put process ID into argblk for fork call (see definition of argblk
 * above).
 */
      sprintf(ppid,"%d",getpid());

/*
 * Create the pipe als standard input for the forked alarmclk process
 * and execute the fork. Restore standard input for the own process
 * afterwards.
 */
      StdIn=dup(0);
      close(0);	
      creat("/pipe",S_IREAD | S_IWRITE);

      if((alarmpid=os9exec(os9forkc,"alarmclk",argblk,environ,0,0,3)) < 0)
      {
#ifdef DEBUG
         fprintf(stderr,"alarm(): start of alarm clock failed\n");
#endif
         return(ALRMERR);
      }
      fildes=dup(0);
      close(0);
      dup(StdIn);
      close(StdIn);
   }
/*
 * For all subsequent calls, calculate the return value. Make sure
 * it cannot be negative. 0 is returned if the previous call expired
 * its given time (and the signal was therefore sent).
 */
   else
   {
      time(&timenow);
      timeleft=difftime(timenow,timeold);
      timeold=timenow;
      timeleft=(slpsec > timeleft) ? slpsec-timeleft : 0;
   }

/*
 * Reset the interrupt flag and write to the pipe.
 * Note that at this point it no longer matters wether the function
 * is interrupted: if the interrupt happens before the old value
 * is written to the pipe, the interrupted call will write only after
 * the interrupting one has finished, which means that the updated
 * value will be written twice - the alarmclock process will handle
 * this correctly. If the interrupt happens, however, after the write,
 * the old value will be written before the new one, which simply
 * corresponds to two non-interrupting calls happening very close
 * to each other.
 */
   alrmlock=0;
   slpsec=secval;
   if((status=write(fildes,&slpsec,PIPDAT)) == PIPDAT)
      return(timeleft);
#ifdef DEBUG
   if((status < 0) && (errno == E_WRITE))
      fprintf(stderr,"alarm(): pipe full\n");
   else
      fprintf(stderr,"alarm(): pipe write error\n");
#endif
   return(ALRMERR);
}
      
/*
 * Function to kill the alarmclk process that was started
 * by the alarm() call.
 *
 * Usage: int alrmkill()
 *
 * The alarmclk process is killed by sending a "system abort"-
 * signal to the process. alrmkill returns the return value of
 * the "kill" function.
 */

int alrmkill()
{
   int status,waitid;

   status=kill(alarmpid,SIGKILL);
   if(status != -1)
   {
   /*
    * Wait for the alarmclk process to finish
    */
      do
      {
         waitid=wait(0);
      } while(waitid > 0 && waitid != alarmpid);
      alarmpid=-1;
   }
   return(status);
}
