/*
 * Version 14/04/89
 * 04/04/89 raw: change mechanism for alarm(0) call
 * 14/04/89 raw: put error output under #ifdef DEBUG
 * 
 * Alarmclock process routine of the "alarm-package" for OS-9.
 * The package simulates the UNIX alarm() function for OS-9.
 * It consists out of
 * - this alarmclock process;
 * - the alarm() function.
 *
 * Note: an executable module of "alarmclk" must be present in the 
 *       working execution directory for the alarm-package to work.  
 *
 * The user calls alarm(sec) from his program, as he would for the
 * UNIX routine. alarm() takes the input value and sends it over an
 * unnamed pipe to the alarmclock process; for details on this see
 * the description of the alarm() function.
 * This process reads the alarm() input value from the pipe, sleeps
 * for this amount of time and sends the SIGALRM signal to the calling
 * process afterwards. SIGALRM can be defined in the signal.h file
 * by the user; if it is not, the (UNIX) default value 14 will be 
 * taken. A 0 input value will cause the process to wait indefinitely
 * for the arrival of a new input value, as desired for the alarm(0) 
 * call. 
 * The pipe is constantly - even during the sleep - monitored for the 
 * arrival of new data. In this case, the new data will be read, the old 
 * values forgotten, and the process continues as requested by the new
 * values. This corresponds to the characteristics of the UNIX alarm()
 * function, where a repeated call to the function will override the 
 * values of the old calls.
 * The pipe monitoring uses a facility of OS-9 that sends a signal to the
 * calling process when data arrives at a device (the pipe in this case).
 *
 * Author: R. Wilcke (ESRF, Grenoble)
 */

/*
 * The following define is needed to make the alarmclk process work
 * for the network compiler. It can be ignored or removed otherwise.
 */
#define OS9 OSK

#include <stdio.h>
#include <modes.h>
#include <errno.h>
#include <signal.h>

/* define SIGALRM if it has not been defined in the signal.h file */
#ifndef SIGALRM
#define SIGALRM 14
#endif

#define SIGIO 16

int readwait;
unsigned slpsec;

main(argc,argv)
int argc;
char *argv[];
{
#define PIPDAT 4

   int ppid,status;
   unsigned sec;
   static int fildes=0;
   int readsign();

/*
 * Input monitoring of the pipe is done with signals. Set up the
 * signal handler here.
 */
   intercept(readsign);
/*
 * Get the parent process ID (= process that called alarm()) from the
 * input arguments, in order to know where to send the signal.
 */
   ppid=atoi(argv[1]);

/* Endless loop: read pipe, sleep specified time, send signal;
 *               go to read if interrupted by a signal.
 * Loop is terminated by EOF on pipe; or by error.
 */
   for(;;)
   {
/*
 * Disable the input monitor, then read the pipe.
 */
      _ss_rel(fildes);
      if((status=read(fildes,&sec,PIPDAT)) < PIPDAT)
/*
 * Check for read errors (status < 0 or less than PIPDAT bytes read)
 * and end-of-file (0 bytes read).
 */ 
      {
         if(status<0)
         {
#ifdef DEBUG
            fprintf(stderr,
               "alarmclk process %d: pipe read error = %d, exit\n",
               getpid(),status);
#endif
            exit(1);
         } 
         else
         {
            if(status==0)
               exit(0);

#ifdef DEBUG
            fprintf(stderr,
               "alarmclk process %d: EOF on pipe after %d bytes, exit\n",
                  getpid(),status);
#endif
            exit(1);
         }
      }
/*
 * For alarm(0) call, do not sleep, but wait at read statement.
 */
      if(sec == 0)
         continue;
      slpsec=sec;
      readwait=0;

/*
 * (Re-)Enable the input monitoring.
 */
      if((status=_ss_ssig(fildes,SIGIO)) == -1)
      {
#ifdef DEBUG
         fprintf(stderr,"alarmclk process %d: signal setup error %x,exit\n",
            getpid(),errno);
#endif
         exit(1);
      }

/*
 * If there is still unread data in the pipe, continue reading and forget
 * old values.
 */   
      if(readwait == 0)
      {
/*
 * Go to sleep.
 */
         sleep(slpsec);
/*
 * Find out if sleep was terminated by a signal: then read new data;
 * or expired naturally: then send signal.
 */
         if(readwait == 0)
         {
            if((status=kill(ppid,SIGALRM)) == -1)
            {
#ifdef DEBUG
               fprintf(stderr,
                  "alarmclk process %d: kill send error = %x, exit\n",
                  getpid(),errno);
#endif
               exit(1);
            }
         }
      }
   }
} 

/*
 * Signal handler function. Simply sets the readwait flag.
 * Note: in the (rare) case that the signal arrives just between the
 * readwait test and the sleep call, the sleep can no longer be
 * prevented, but it will last only 1 second.
 */ 
readsign(signum)
int signum;
{
   if(signum == SIGIO)
   {
      readwait=1;
      slpsec=1;
   }
}
